import inspect
import unittest

import pymzm
import minizinc

import equiv
from functools import partial


class TestExpression(unittest.TestCase):
    def setUp(self):
        self.solver = minizinc.Solver.lookup("gecode")
        self.val_min = -100
        self.val_max = 100

        self.model = pymzm.Model()
        self.x = self.model.add_variable("x", val_min=self.val_min, val_max=self.val_max)
        self.xs = self.model.add_variables("xs", range(10), val_min=self.val_min, val_max=self.val_max)
        self.y = self.model.add_variable("y", vtype=pymzm.Variable.VTYPE_BOOL)

    def _operator_case_single(self, func, positive_only=False, val_max: int=None):
        model = pymzm.Model()
        val_min = 0 if positive_only else self.val_min
        val_max = self.val_max if val_max is None else val_max
        
        x = model.add_variable("x", vtype=pymzm.Variable.VTYPE_INTEGER, val_min=val_min, val_max=val_max)
        model.add_constraint(func(x))
        model.set_solve_criteria(pymzm.SOLVE_SATISFY)
        model.generate()

        sols = []
        for i in range(val_min, val_max + 1):
            try:
                if (func(i)):
                    sols.append(i)
            except ZeroDivisionError:
                pass
        #if (len(sols) == 0): print(inspect.getsource(func))

        results = minizinc.Instance(self.solver, model).solve(all_solutions=True)

        self.assertTrue(results.solution is not None)
        self.assertEqual(len(sols), len(results), f"{sols}, {results}")
        for i in range(len(results)):
            self.assertTrue(func(results[i, "x"]))
    
    def _operator_case_multiple(self, func_expr, method_mz, method_py=None, var_count: int=5, is_solveable=True):
        model = pymzm.Model()
        
        xs = model.add_variables("x", range(var_count), vtype=pymzm.Variable.VTYPE_INTEGER, val_min=self.val_min, val_max=self.val_max)

        func_mz = partial(func_expr, method_mz)
        func_py = partial(func_expr, method_py) if method_py is not None else None

        model.add_constraint(func_mz(xs))
        model.set_solve_criteria(pymzm.SOLVE_SATISFY)
        model.generate()

        result = minizinc.Instance(self.solver, model).solve(all_solutions=False)

        if (is_solveable):
            self.assertTrue(result.solution is not None)
            if (func_py is None):
                func_py = func_mz
            answ = func_py([result[f"x_{i}"] for i in range(var_count)])
            self.assertIsInstance(answ, bool) # for some reason non bools evaluate to true
            self.assertTrue(answ)

    def test_native_sum(self):
        # Valid
        sum(x for x in self.xs)
        sum([x for x in self.xs])
        sum(self.xs)
        #sum(self.xs[0], self.xs[1], self.y)

        self._operator_case_multiple(lambda f, xs: 0 + f(xs) == 27, sum, var_count=5)
        self._operator_case_multiple(lambda f, xs: 0 - f(xs) == 31, sum, var_count=5)
        self._operator_case_multiple(lambda f, xs: 5 * f(xs) == 30, sum, var_count=5)
        self._operator_case_multiple(lambda f, xs: f(xs) * f(xs) == 25, sum, var_count=2)
        self._operator_case_multiple(lambda f, xs: - f(xs) == 14, sum, var_count=5)

        expr = sum(self.xs)
        self.assertIsInstance(expr, pymzm.Expression)
        
        # Invalid
        #self.assertRaises(pymzm.PymzmValueIsNotExpression,
        #    sum,
        #    ["test", self.xs[0]])
        #self.assertRaises(pymzm.PymzmValueIsNotExpression,
        #    sum,
        #    [[self.xs[1]], self.xs[0]])
        #self.assertRaises(pymzm.PymzmValueIsNotExpression,
        #    sum,
        #    self.xs[0])
        #self.assertRaises(pymzm.PymzmValueIsNotExpression,
        #    sum,
        #    "test")
        #self.assertRaises(pymzm.PymzmNoValues,
        #    sum,
        #    [])

    def test_sum(self):
        # Valid
        pymzm.Expression.sum(x for x in self.xs)
        pymzm.Expression.sum([x for x in self.xs])
        pymzm.Expression.sum(self.xs)
        #pymzm.Expression.sum(self.xs[0], self.xs[1], self.y)

        self._operator_case_multiple(lambda f, xs: 0 + f(xs) == 27, pymzm.Expression.sum, sum, var_count=5)
        self._operator_case_multiple(lambda f, xs: 0 - f(xs) == 31, pymzm.Expression.sum, sum, var_count=5)
        self._operator_case_multiple(lambda f, xs: 5 * f(xs) == 30, pymzm.Expression.sum, sum, var_count=5)
        self._operator_case_multiple(lambda f, xs: f(xs) * f(xs) == 25, pymzm.Expression.sum, sum, var_count=2)
        self._operator_case_multiple(lambda f, xs: - f(xs) == 14, pymzm.Expression.sum, sum, var_count=5)

        expr = pymzm.Expression.sum(self.xs)
        self.assertIsInstance(expr, pymzm.Expression)
        
        # Invalid
        self.assertRaises(pymzm.PymzmValueIsNotExpression,
            pymzm.Expression.sum,
            ["test", self.xs[0]])
        self.assertRaises(pymzm.PymzmValueIsNotExpression,
            pymzm.Expression.sum,
            [[self.xs[1]], self.xs[0]])
        self.assertRaises(pymzm.PymzmValueIsNotExpression,
            pymzm.Expression.sum,
            self.xs[0])
        self.assertRaises(pymzm.PymzmValueIsNotExpression,
            pymzm.Expression.sum,
            "test")
        self.assertRaises(pymzm.PymzmNoValues,
            pymzm.Expression.sum,
            [])

    def test_product(self):
        # Valid
        pymzm.Expression.product(x for x in self.xs)
        pymzm.Expression.product([x for x in self.xs])
        pymzm.Expression.product(self.xs)
        #pymzm.Expression.product(self.xs[0], self.xs[1], self.y)

        self._operator_case_multiple(lambda f, xs: 0 + f(xs) == 2*1*1, pymzm.Expression.product, equiv.Expression.product, var_count=3)
        self._operator_case_multiple(lambda f, xs: 0 - f(xs) == 5*6*(-3), pymzm.Expression.product, equiv.Expression.product, var_count=3)
        self._operator_case_multiple(lambda f, xs: 5 * f(xs) == 11*5*1, pymzm.Expression.product, equiv.Expression.product, var_count=3)
        self._operator_case_multiple(lambda f, xs: f(xs) + f(xs) == 2*11*6*2, pymzm.Expression.product, equiv.Expression.product, var_count=3)
        self._operator_case_multiple(lambda f, xs: - f(xs) == 5*5*5, pymzm.Expression.product, equiv.Expression.product, var_count=3)

        expr = pymzm.Expression.product(self.xs)
        self.assertIsInstance(expr, pymzm.Expression)
        
        # Invalid
        self.assertRaises(pymzm.PymzmValueIsNotExpression,
            pymzm.Expression.product,
            ["test", self.xs[0]])
        self.assertRaises(pymzm.PymzmValueIsNotExpression,
            pymzm.Expression.product,
            [[self.xs[1]], self.xs[0]])
        self.assertRaises(pymzm.PymzmValueIsNotExpression,
            pymzm.Expression.product,
            self.xs[0])
        self.assertRaises(pymzm.PymzmValueIsNotExpression,
            pymzm.Expression.product,
            "test")
        self.assertRaises(pymzm.PymzmNoValues,
            pymzm.Expression.product,
            [])

    def test_min(self):
        # Valid
        pymzm.Expression.min(x for x in self.xs)
        pymzm.Expression.min([x for x in self.xs])
        pymzm.Expression.min(self.xs)

        self._operator_case_multiple(lambda f, xs: f([*xs, 5]) == 5, pymzm.Expression.min, min, var_count=5)
        self._operator_case_multiple(lambda f, xs: f([*xs, 2]) == 5, pymzm.Expression.min, min, var_count=5, is_solveable=False)
        self._operator_case_multiple(lambda f, xs: f([*xs, 7]) == 5, pymzm.Expression.min, min, var_count=5)
        self._operator_case_multiple(lambda f, xs: f([*xs, -5]) == 5, pymzm.Expression.min, min, var_count=5, is_solveable=False)

        expr = pymzm.Expression.min(self.xs)
        self.assertIsInstance(expr, pymzm.Expression)
        
        # Invalid
        self.assertRaises(pymzm.PymzmValueIsNotExpression,
            pymzm.Expression.min,
            ["test", self.xs[0]])
        self.assertRaises(pymzm.PymzmValueIsNotExpression,
            pymzm.Expression.min,
            [[self.xs[1]], self.xs[0]])
        self.assertRaises(pymzm.PymzmValueIsNotExpression,
            pymzm.Expression.min,
            self.xs[0])
        self.assertRaises(pymzm.PymzmValueIsNotExpression,
            pymzm.Expression.min,
            "test")
        self.assertRaises(pymzm.PymzmNoValues,
            pymzm.Expression.min,
            [])

    def test_max(self):
        # Valid
        pymzm.Expression.max(x for x in self.xs)
        pymzm.Expression.max([x for x in self.xs])
        pymzm.Expression.max(self.xs)

        self._operator_case_multiple(lambda f, xs: f([*xs, 5]) == 5, pymzm.Expression.max, max, var_count=5)
        self._operator_case_multiple(lambda f, xs: f([*xs, 2]) == 5, pymzm.Expression.max, max, var_count=5, is_solveable=False)
        self._operator_case_multiple(lambda f, xs: f([*xs, 7]) == 5, pymzm.Expression.max, max, var_count=5, is_solveable=False)
        self._operator_case_multiple(lambda f, xs: f([*xs, -5]) == 5, pymzm.Expression.max, max, var_count=5, is_solveable=False)

        expr = pymzm.Expression.max(self.xs)
        self.assertIsInstance(expr, pymzm.Expression)
        
        # Invalid
        self.assertRaises(pymzm.PymzmValueIsNotExpression,
            pymzm.Expression.max,
            ["test", self.xs[0]])
        self.assertRaises(pymzm.PymzmValueIsNotExpression,
            pymzm.Expression.max,
            [[self.xs[1]], self.xs[0]])
        self.assertRaises(pymzm.PymzmValueIsNotExpression,
            pymzm.Expression.max,
            self.xs[0])
        self.assertRaises(pymzm.PymzmValueIsNotExpression,
            pymzm.Expression.max,
            "test")
        self.assertRaises(pymzm.PymzmNoValues,
            pymzm.Expression.max,
            [])

    def test_funcs_pow(self):
        self._operator_case_single(lambda x: pow(x, 2) == 81)
        self._operator_case_single(lambda x: pow(x, 2) == 5)
        self._operator_case_single(lambda x: pow(x, 3) == 27)
        self._operator_case_single(lambda x: pow(2, x) == 64, val_max=10)
        self._operator_case_single(lambda x: x ** 2 == 81)
        self._operator_case_single(lambda x: x ** 2 == 5)
        self._operator_case_single(lambda x: x ** 3 == 27)
        self._operator_case_single(lambda x: 3 ** x == 243, val_max=10)
        
        expr = pymzm.Expression.product(self.xs)
        self.assertIsInstance(expr, pymzm.Expression)

    def test_funcs_abs(self):
        self._operator_case_single(lambda x: abs(-x) == 81)
        self._operator_case_single(lambda x: -abs(x) == 5)
        self._operator_case_single(lambda x: abs(x) == 27)
        self._operator_case_single(lambda x: abs(x) == -27)

        expr = pymzm.Expression.product(self.xs)
        self.assertIsInstance(expr, pymzm.Expression)

    def test_add(self):
        self._operator_case_single(lambda x: x + x == 4)
        self._operator_case_single(lambda x: 5 + x == 4)
        self._operator_case_single(lambda x: x + 4 == 4)
        
        self.assertIsInstance(self.x + self.x, pymzm.Expression)
        self.assertIsInstance(5 + self.x, pymzm.Expression)
        self.assertIsInstance(self.x + 5, pymzm.Expression)

    def test_sub(self):
        self._operator_case_single(lambda x: x - x == 4)
        self._operator_case_single(lambda x: 5 - x == 4)
        self._operator_case_single(lambda x: x - 4 == 4)
        
        self.assertIsInstance(self.x - self.x, pymzm.Expression)
        self.assertIsInstance(5 - self.x, pymzm.Expression)
        self.assertIsInstance(self.x - 5, pymzm.Expression)

    def test_mul(self):
        self._operator_case_single(lambda x: x * x == 16)
        self._operator_case_single(lambda x: 5 * x == 10)
        self._operator_case_single(lambda x: x * 4 == 4)
        
        self.assertIsInstance(self.x * self.x, pymzm.Expression)
        self.assertIsInstance(5 * self.x, pymzm.Expression)
        self.assertIsInstance(self.x * 5, pymzm.Expression)

    def test_truediv(self):
        self._operator_case_single(lambda x: x / x == 1)
        self._operator_case_single(lambda x: 20 / x == 10)
        self._operator_case_single(lambda x: x / 2 == 4)
        
        self.assertIsInstance(self.x / self.x, pymzm.Expression)
        self.assertIsInstance(2 / self.x, pymzm.Expression)
        self.assertIsInstance(self.x / 2, pymzm.Expression)

    def test_floordiv(self):
        self._operator_case_single(lambda x: x // x == 1)
        self._operator_case_single(lambda x: 30 // x == 10)
        self._operator_case_single(lambda x: x // 4 == 4)
        
        self.assertIsInstance(self.x // self.x, pymzm.Expression)
        self.assertIsInstance(5 // self.x, pymzm.Expression)
        self.assertIsInstance(self.x // 5, pymzm.Expression)

    def test_mod(self):
        self._operator_case_single(lambda x: x % x == 0, positive_only=True)
        self._operator_case_single(lambda x: 5 % x == 2, positive_only=True)
        self._operator_case_single(lambda x: x % 4 == 3, positive_only=True)
        
        self.assertIsInstance(self.x % self.x, pymzm.Expression)
        self.assertIsInstance(5 % self.x, pymzm.Expression)
        self.assertIsInstance(self.x % 5, pymzm.Expression)

    def test_neg(self):
        self._operator_case_single(lambda x: - x == 16)
        self._operator_case_single(lambda x: - x == -3)
        
        self.assertIsInstance(-self.x, pymzm.Expression)

    def test_ifthenelse(self):
        # Invalid
        self.assertRaises(pymzm.PymzmValueIsNotCondition, 
            pymzm.Expression.ifthenelse,
            self.xs[0] + self.xs[1], self.xs[3], self.xs[4])
        self.assertRaises(pymzm.PymzmValueIsNotExpression, 
            pymzm.Expression.ifthenelse,
            self.y, "test", self.xs[4])
        self.assertRaises(pymzm.PymzmValueIsNotExpression, 
            pymzm.Expression.ifthenelse,
            self.y, self.xs[4], "test")
        self.assertRaises(pymzm.PymzmValueIsNotExpression, 
            pymzm.Expression.ifthenelse,
            self.y > self.xs[1], self.xs[4], "test")
        
        # Valid
        pymzm.Expression.ifthenelse(self.y, self.y, self.xs[4])
        pymzm.Expression.ifthenelse(self.xs[0] >= self.xs[1], self.y, self.xs[4])
        pymzm.Expression.ifthenelse(self.xs[1] >= self.y, self.y, self.xs[4])

        expr = pymzm.Expression.ifthenelse(self.xs[1] >= self.y, self.y, self.xs[4])
        self.assertIsInstance(expr, pymzm.Expression)

if __name__ == "__main__":
    t = TestExpression()
    t.setUpClass()
    t.setUp()
    t.test_ifthenelse()

    unittest.main()



import minizinc
from typing import List, Tuple

from .exceptions import *
from .variable import *
from .constraint import *
from .expression import *
from .constant import *

SOLVE_MAXIMIZE = "maximize"
SOLVE_MINIMIZE = "minimize"
SOLVE_SATISFY = "satisfy"

class RestartStrategy:
    def __init__(self, restart_type: str, scale: int):
        self.restart_type = restart_type
        self.scale = scale

    def __str__(self):
        return f"{self.restart_type}({self.scale})"

class RestartConstant(RestartStrategy):
    def __init__(self, scale):
        super().__init__("restart_constant", scale)

class RestartLinear(RestartStrategy):
    def __init__(self, scale):
        super().__init__("restart_linear", scale)

class RestartGeometric(RestartStrategy):
    def __init__(self, base, scale):
        self.base = base
        super().__init__("restart_geometric", scale)

    def __str__(self):
        return f"{self.restart_type}({self.base}, {self.scale})"

class RestartLuby(RestartStrategy):
    def __init__(self, scale):
        super().__init__("restart_luby", scale)



class SeqSearch:
    def __init__(self, search_annotations: list["SearchAnnotation"]):
        self.search_annotations = search_annotations
        if (not all(isinstance(sa, SearchAnnotation) for sa in search_annotations)):
            raise PymzmInvalidSearchAnnotation()
        
    def __str__(self):
        return f"seq_search([{','.join(str(sa) for sa in self.search_annotations)}])"

class SearchAnnotation:
    def __init__(self, search_type: str, variables: List["Variable"], varchoice: str, valchoice: str):
        self.search_type = search_type
        self.variables = variables

        self.varchoice = varchoice
        if (not self.varchoice in AnnotationVariableChoice.VARCHOICES):
            raise PymzmInvalidVarchoiceAnnotation("varchoice")
        
        self.valchoice = valchoice
        if (not self.valchoice in AnnotationValueChoice.VALCHOICES):
            raise PymzmInvalidValchoiceAnnotation("valchoice")
        
    def __str__(self):
        return f"{self.search_type}({self.variables}, {self.varchoice}, {self.valchoice})"

class IntSearch(SearchAnnotation):
    def __init__(self, variables: List["Variable"], varchoice: str, valchoice: str):
        if (not all(isinstance(v, Variable) for v in variables)):
            raise PymzmInvalidVariableError("variables", "Atleast one variable is not an integer.")
        
        super().__init__("int_search", variables, varchoice, valchoice)

class BoolSearch(SearchAnnotation):
    def __init__(self, variables: List["VariableBool"], varchoice: str, valchoice: str):
        if (not all(isinstance(v, VariableBool) for v in variables)):
            raise PymzmInvalidVariableError("variables", "Atleast one variable is not a bool.")
        
        super().__init__("bool_search", variables, varchoice, valchoice)

class SetSearch(SearchAnnotation):
    def __init__(self, variables: List["Expression"], varchoice: str, valchoice: str):
        if (not all(isinstance(v, Variable) or not all(v.vtype == Variable.VTYPE_SET) for v in variables)):
            raise PymzmInvalidVariableError("variables", "Atleast one variable is not a set.")
        
        super().__init__("set_search", variables, varchoice, valchoice)

class FloatSearch(SearchAnnotation):
    def __init__(self, variables: List["Expression"], precision: float, varchoice: str, valchoice: str):
        raise NotImplementedError()
    
class AnnotationVariableChoice:
    VARCHOICES = [
        VARCHOICE_INPUT_ORDER, # Choose variables in the order they appear in vars.
        VARCHOICE_FIRST_FAIL, # Choose the variable with the smallest domain.
        VARCHOICE_ANTI_FIRST_FAIL, # Choose the variable with the largest domain
        VARCHOICE_SMALLEST, #Choose the variable with the smallest value in its domain.
        VARCHOICE_LARGEST, #Choose the variable with the largest value in its domain.
        VARCHOICE_OCCURRENCE, # Choose the variable with the largest number of attached constraints.
        VARCHOICE_MOST_CONSTRAINED, # Choose the variable with the smallest domain, breaking ties using the number of constraints.
        VARCHOICE_MAX_REGRET, # Choose the variable with the largest difference between the two smallest values in its domain.
        VARCHOICE_DOM_W_DEG, # Choose the variable with the smallest value of domain size divided by weighted degree, where the weighted degree is the number of times the variables been in a constraint which failed
    ] = [
        "input_order",
        "first_fail",
        "anti_first_fail",
        "smallest",
        "largest",
        "occurrence",
        "most_constrained",
        "max_regret",
        "dom_w_deg",
    ]

class AnnotationValueChoice:
    VALCHOICES = [
        VALCHOICE_INDOMAIN_MIN, # Assign the smallest value in the variable’s domain.
        VALCHOICE_INDOMAIN_MAX, # Assign the largest value in the variable’s domain.
        VALCHOICE_INDOMAIN_MIDDLE, # Assign the value in the variable’s domain closest to the mean of its current bounds.
        VALCHOICE_INDOMAIN_MEDIAN, # Assign the middle value in the variable’s domain.
        VALCHOICE_INDOMAIN, # Nondeterministically assign values to the variable in ascending order.
        VALCHOICE_INDOMAIN_RANDOM, # Assign a random value from the variable’s domain.
        VALCHOICE_INDOMAIN_SPLIT, # Bisect the variable’s domain, excluding the upper half first.
        VALCHOICE_INDOMAIN_REVERSE_SPLIT, # Bisect the variable’s domain, excluding the lower half first.
        VALCHOICE_INDOMAIN_INTERVAL, # If the variable’s domain consists of several contiguous intervals, reduce the domain to the first interval. Otherwise just split the variable’s domain.
    ] = [
        "indomain_min",
        "indomain_max",
        "indomain_middle",
        "indomain_median",
        "indomain",
        "indomain_random",
        "indomain_split",
        "indomain_reverse_split",
        "indomain_interval",
    ]


class Model(minizinc.Model):
    def __init__(self):
        self.constants = []
        self.variables = []
        self.constraints = []
        self.solve_criteria = None
        self.solve_expression = None
        self.solve_method = None
        self.model_mzn_str = None

        self.global_constraints = set()

        super().__init__()

    def set_solve_criteria(self, criteria: str, expr: Expression=None):
        self.solve_criteria = criteria
        self.solve_expression = expr

        if (criteria == SOLVE_MAXIMIZE or criteria == SOLVE_MINIMIZE):
            assert expr is not None
        elif (criteria == SOLVE_SATISFY):
            assert expr is None
        else:
            raise Exception(f"Invalid solve criteria: {criteria}")

    def set_solve_method(self, method: SearchAnnotation, restart_strategy: RestartStrategy=None):
        assert isinstance(method, (SeqSearch, SearchAnnotation))
        if (restart_strategy is not None):
            assert isinstance(restart_strategy, RestartStrategy)
        self.solve_method = method
        self.restart_strategy = restart_strategy

    def add_constant(self, name: str, value, vtype=Variable.VTYPE_INTEGER):
        constant = Constant(name, value, vtype)
        self.constants.append(constant)
        return constant

    def add_variable(self, name: str, vtype: int=Variable.VTYPE_INTEGER, val_min: int=None, val_max: int=None, domain: set=None):
        variable = Variable(name, vtype, val_min, val_max, domain)
        self.variables.append(variable)
        return variable
    
    def add_variables(self, name: str, indices: List[Tuple[int]], vtype: int=Variable.VTYPE_INTEGER, val_min: int=None, val_max: int=None, domains: set=None) -> ValueDict:

        # Domain
        if (domains is None):
            domains = {}
        elif (type(domains) is set):
            domains = {idx: domains for idx in indices}
        elif (type(domains) is list):
            assert len(domains) == len(indices)
            domains = {idx: domains[i] for i, idx in enumerate(indices)}
        elif (type(domains) is dict):
            assert len(domains) == len(indices)
            assert set(domains.keys()) == set(indices)
            domains = {idx: domains[idx] for idx in indices}

        variables = ValueDict()
        for idx in indices:
            idx_str = str(idx).replace(", ", "_").replace("(", "").replace(")", "").replace("'", "").replace("-", "_")
            variable = Variable(f"{name}_{idx_str}", vtype, val_min, val_max, domains.get(idx, None))
            self.variables.append(variable)
            variables[idx] = variable

        return variables

    def add_constraint(self, constraint: ExpressionBool, is_redundant=False):
        if (isinstance(constraint, Constraint)):
            constraint.is_redundant = is_redundant
            if (constraint.ctype != Constraint.CTYPE_NORMAL):
                self.global_constraints.add(constraint.ctype)

        elif (isinstance(constraint, ExpressionBool)):
            constraint = Constraint(constraint.name, is_redundant=is_redundant)

        else:
            raise Exception("invalid constraint type")

        self.constraints.append(constraint)
        return constraint

    def add_constraints(self, constraints: List[Constraint], is_redundant=False):
        constraints = list(constraints)
        assert all(isinstance(constraint, (Constraint, Expression, str, bool)) for constraint in constraints)
        for constraint in constraints:
            self.add_constraint(constraint, is_redundant=is_redundant)

    def generate(self, debug=False):
        self.model_mzn_str = ""
        for gconst in self.global_constraints:
            self.model_mzn_str += f'include \"{gconst}.mzn\";\n'

        self.model_mzn_str += "".join(a._to_mz() for a in self.constants + self.variables + self.constraints)
        
        assert self.solve_criteria is not None
        _solve_method_str = ""
        if (self.solve_method is not None):
            _solve_method_str += f":: {self.solve_method}\n"
            
            if (self.restart_strategy is not None):
                _solve_method_str += f"      :: {self.restart_strategy}\n "

        _solve_method_str += f"{self.solve_criteria}"

        if (self.solve_expression is not None):
            _solve_method_str += f" {self.solve_expression}"
            
        self.model_mzn_str += f"solve {_solve_method_str};\n"

        self.add_string(self.model_mzn_str)
        if (debug):
            print(self.model_mzn_str)

    def write(self, fn: str):
        if (self.model_mzn_str is None):
            self.generate()
        with open(fn, "w") as f:
            f.write(self.model_mzn_str)



from typing import List
from collections.abc import Iterable

from .exceptions import *

class Expression:
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name
    
    def __repr__(self):
        return self.name

    def __bool__(self):
        return False

    @staticmethod
    def ifthenelse(condition: "ExpressionBool", expr1: "Expression", expr2: "Expression") -> "Expression":
        """ifelse: if (condition) then expr1 else expr2:

        Args:
            condition (Expression): condition of expression
            expr1 (Expression): expression if condition
            expr2 (Expression): expression else

        Returns:
            Expression: the main if then else expression 
        """
        if (not isinstance(condition, (ExpressionBool, bool))):
            raise PymzmValueIsNotCondition("condition", condition)
        
        if (not isinstance(expr1, (Expression, int, float))):
            raise PymzmValueIsNotExpression("expr1", expr1)
        
        if (not isinstance(expr2, (Expression, int, float))):
            raise PymzmValueIsNotExpression("expr2", expr2)
        
        return Expression(f"(if {condition} then {expr1} else {expr2} endif)")

    @staticmethod
    def sum(exprs: List["Expression"]) -> "Expression":
        if (not isinstance(exprs, Iterable)):
            raise PymzmValueIsNotExpression("exprs", exprs)
        
        #exprs = list(exprs)
        if (not len(exprs)):
            return Expression("0")
            raise PymzmNoValues("exprs")

        for expr in exprs:
            if (not isinstance(expr, (Expression, int, float))):
                raise PymzmValueIsNotExpression("exprs", exprs)
            
        return Expression._func("sum", [exprs])
    
    @staticmethod
    def product(exprs) -> "Expression":
        if (not isinstance(exprs, Iterable)):
            raise PymzmValueIsNotExpression("exprs", exprs)
        
        exprs = list(exprs)
        if (not len(exprs)):
            raise PymzmNoValues("exprs")

        for expr in exprs:
            if (not isinstance(expr, (Expression, int, float))):
                raise PymzmValueIsNotExpression("exprs", exprs)
            
        return Expression._func("product", [exprs])

    @staticmethod
    def min(exprs: List["Expression"]) -> "Expression":
        if (not isinstance(exprs, Iterable)):
            raise PymzmValueIsNotExpression("exprs", exprs)
        
        exprs = list(exprs)
        if (not len(exprs)):
            raise PymzmNoValues("exprs")

        for expr in exprs:
            if (not isinstance(expr, (Expression, int, float))):
                raise PymzmValueIsNotExpression("exprs", exprs)
            
        return Expression._func("min", [exprs])

    @staticmethod
    def max(exprs: List["Expression"]) -> "Expression":
        if (not isinstance(exprs, Iterable)):
            raise PymzmValueIsNotExpression("exprs", exprs)
        
        exprs = list(exprs)
        if (not len(exprs)):
            raise PymzmNoValues("exprs")

        for expr in exprs:
            if (not isinstance(expr, (Expression, int, float))):
                raise PymzmValueIsNotExpression("exprs", exprs)
            
        return Expression._func("max", [exprs])

    @classmethod
    def _operator(cls, symbol: str, exprs):
        exprs = [expr.name if isinstance(expr, Expression) else expr for expr in exprs]
        out = f" {symbol} ".join(str(a) for a in exprs)
        out = f"({out})"

        return cls(out)

    @classmethod
    def _func(cls, func_symbol: str, exprs):
        exprs2 = []
        for expr in exprs:
            if (isinstance(expr, Expression)):
                exprs2.append(expr.name)
            else:
                exprs2.append(expr)
        out = f", ".join(str(a) for a in exprs2)
        return cls(f"{func_symbol}({out})")

    @staticmethod
    def OR(exprs: List["ExpressionBool"]) -> "ExpressionBool":
        for expr in exprs:
            if (not isinstance(expr, ExpressionBool)):
                raise PymzmValueIsNotCondition("exprs", expr)

        return ExpressionBool._operator("\/", exprs)
    
    @staticmethod
    def AND(exprs: List["ExpressionBool"]) -> "ExpressionBool":
        for expr in exprs:
            if (not isinstance(expr, ExpressionBool)):
                raise PymzmValueIsNotCondition("exprs", expr)

        return ExpressionBool._operator("/\\", exprs)
    
    @staticmethod
    def onlyIf(exprs: List["ExpressionBool"]) -> "ExpressionBool":
        for expr in exprs:
            if (not isinstance(expr, ExpressionBool)):
                raise PymzmValueIsNotCondition("exprs", expr)

        return ExpressionBool._operator("<-", exprs)
    
    @staticmethod
    def implies(exprs: List["ExpressionBool"]) -> "ExpressionBool":
        for expr in exprs:
            if (not isinstance(expr, ExpressionBool)):
                raise PymzmValueIsNotCondition("exprs", expr)

        return ExpressionBool._operator("->", exprs)
    
    @staticmethod
    def iff(exprs: List["ExpressionBool"]) -> "ExpressionBool":
        # <->
        for expr in exprs:
            if (not isinstance(expr, ExpressionBool)):
                raise PymzmValueIsNotCondition("exprs", expr)

        return ExpressionBool._operator("<->", exprs)
    
    @staticmethod
    def xor(exprs: List["ExpressionBool"]) -> "ExpressionBool":
        for expr in exprs:
            if (not isinstance(expr, ExpressionBool)):
                raise PymzmValueIsNotCondition("exprs", expr)

        return ExpressionBool._operator("xor", exprs)
    
    @staticmethod
    def NOT(expr: "ExpressionBool") -> "ExpressionBool":
        if (not isinstance(expr, ExpressionBool)):
            raise PymzmValueIsNotCondition("exprs", expr)

        return ExpressionBool._func("not", [expr])

    def __add__(self, other: "Expression") -> "Expression":
        if (not isinstance(other, (int, float, Expression))):
            raise PymzmValueIsNotExpression("other", other)

        return Expression._operator("+", [self, other])
    
    def __radd__(self, other: "Expression") -> "Expression":
        if (not isinstance(other, (int, float, Expression))):
            raise PymzmValueIsNotExpression("other", other)

        return Expression._operator("+", [other, self])
    
    def __sub__(self, other: "Expression") -> "Expression":
        if (not isinstance(other, (int, float, Expression))):
            raise PymzmValueIsNotExpression("other", other)

        return Expression._operator("-", [self, other])
    
    def __rsub__(self, other: "Expression") -> "Expression":
        if (not isinstance(other, (int, float, Expression))):
            raise PymzmValueIsNotExpression("other", other)

        return Expression._operator("-", [other, self])
    
    def __mul__(self, other: "Expression") -> "Expression":
        if (not isinstance(other, (int, float, Expression))):
            raise PymzmValueIsNotExpression("other", other)

        return Expression._operator("*", [self, other])
    
    def __rmul__(self, other: "Expression") -> "Expression":
        if (not isinstance(other, (int, float, Expression))):
            raise PymzmValueIsNotExpression("other", other)

        return Expression._operator("*", [other, self])
    
    def __truediv__(self, other: "Expression") -> "Expression":
        if (not isinstance(other, (int, float, Expression))):
            raise PymzmValueIsNotExpression("other", other)

        return Expression._operator("/", [self, other])
    
    def __rtruediv__(self, other: "Expression") -> "Expression":
        if (not isinstance(other, (int, float, Expression))):
            raise PymzmValueIsNotExpression("other", other)

        return Expression._operator("/", [other, self])
    
    def __floordiv__(self, other: "Expression") -> "Expression":
        if (not isinstance(other, (int, float, Expression))):
            raise PymzmValueIsNotExpression("other", other)

        return Expression._operator("div", [self, other])
    
    def __rfloordiv__(self, other: "Expression") -> "Expression":
        if (not isinstance(other, (int, float, Expression))):
            raise PymzmValueIsNotExpression("other", other)

        return Expression._operator("div", [other, self])
    
    def __mod__(self, other: "Expression") -> "Expression":
        if (not isinstance(other, (int, Expression))):
            raise PymzmValueIsNotExpression("other", other)

        return Expression._operator("mod", [self, other])
    
    def __rmod__(self, other: "Expression") -> "Expression":
        if (not isinstance(other, (int, Expression))):
            raise PymzmValueIsNotExpression("other", other)

        return Expression._operator("mod", [other, self])
    
    def __neg__(self) -> "Expression":
        return 0 - self
    #def __pos__(self): return Expression._func("+", [self]) TODO: not allowed in minizinc example: +x == v
    
    def __eq__(self, other: "Expression") -> "ExpressionBool":
        if (not isinstance(other, (int, float, Expression))):
            raise PymzmValueIsNotExpression("other", other)

        return ExpressionBool._operator("==", [self, other])
    
    def __ne__(self, other: "Expression") -> "ExpressionBool":
        if (not isinstance(other, (int, float, Expression))):
            raise PymzmValueIsNotExpression("other", other)

        return ExpressionBool._operator("!=", [self, other])
    
    def __lt__(self, other: "Expression") -> "ExpressionBool":
        if (not isinstance(other, (int, float, Expression))):
            raise PymzmValueIsNotExpression("other", other)

        return ExpressionBool._operator("<", [self, other])
    
    def __le__(self, other: "Expression") -> "ExpressionBool":
        if (not isinstance(other, (int, float, Expression))):
            raise PymzmValueIsNotExpression("other", other)

        return ExpressionBool._operator("<=", [self, other])
    
    def __gt__(self, other: "Expression") -> "ExpressionBool":
        if (not isinstance(other, (int, float, Expression))):
            raise PymzmValueIsNotExpression("other", other)

        return ExpressionBool._operator(">", [self, other])
    
    def __ge__(self, other: "Expression") -> "ExpressionBool":
        if (not isinstance(other, (int, float, Expression))):
            raise PymzmValueIsNotExpression("other", other)

        return ExpressionBool._operator(">=", [self, other])
    
    
    def __and__(self, other: "ExpressionBool") -> "ExpressionBool":
        if (not isinstance(other, (bool, Expression))):
            raise PymzmValueIsNotCondition("other", other)

        return Expression.AND([self, other])
    
    def __rand__(self, other: "ExpressionBool") -> "ExpressionBool":
        if (not isinstance(other, (bool, Expression))):
            raise PymzmValueIsNotCondition("other", other)

        return Expression.AND([other, self])
    
    def __or__(self, other: "ExpressionBool") -> "ExpressionBool":
        if (not isinstance(other, (bool, Expression))):
            raise PymzmValueIsNotCondition("other", other)

        return Expression.OR([self, other])
    
    def __ror__(self, other: "ExpressionBool") -> "ExpressionBool":
        if (not isinstance(other, (bool, Expression))):
            raise PymzmValueIsNotCondition("other", other)

        return Expression.OR([other, self])
    
    def __xor__(self, other: "ExpressionBool") -> "ExpressionBool":
        if (not isinstance(other, (bool, Expression))):
            raise PymzmValueIsNotCondition("other", other)

        return Expression.xor([self, other])
    
    def __rxor__(self, other: "ExpressionBool") -> "ExpressionBool":
        if (not isinstance(other, (bool, Expression))):
            raise PymzmValueIsNotCondition("other", other)

        return Expression.xor([other, self])
    
    def __invert__(self) -> "ExpressionBool":
        return Expression.NOT(self)

    def __pow__(self, other: "Expression") -> "Expression":
        if (not isinstance(other, (int, float, Expression))):
            raise PymzmValueIsNotExpression("other", other)

        return Expression._func("pow", [self, other])
    
    def __rpow__(self, other: "Expression") -> "Expression":
        if (not isinstance(other, (int, float, Expression))):
            raise PymzmValueIsNotExpression("other", other)
        
        return Expression._func("pow", [other, self])
    
    def __abs__(self) -> "Expression":
        return Expression._func("abs", [self])
    # TODO: https://www.minizinc.org/doc-2.7.6/en/lib-stdlib-builtins.html
    # arg max, arg min
    # max, min
    # count
    # exp(x)
    # log_x, log_2, log_10, ln
    # trinonometric functions
    # ..and more!

class ExpressionBool(Expression):
    pass



from .variable import *
from .misc import *

import numpy as np

class Constant:
    def __init__(self, name: str, value, vtype=Variable.VTYPE_INTEGER):
        self.name = name
        self.value = value
        if (self.value is None):
            raise Exception("Non-initialized constant is not supported by pymzm.")
        
        self.vtype = vtype        
        if (self.vtype not in [Variable.VTYPE_INTEGER, Variable.VTYPE_BOOL]):
            raise Exception("Invalid vtype for constant. Currently only integer and boolean types are supported")
        
        arr = np.array(value)
        if (arr.shape):
            # Is nd array
            self.shape = arr.shape

        else:
            # is single value
            self.shape = None
        
    def __getitem__(self, other: Expression):
        # TODO boolean expression
        if (self.shape is None):
            raise Exception("Constant cant be indexed as it is a single value")
        
        if (len(self.shape) == 1):
            return Expression(f"{self.name}[{str(other)} + 1]")
        
        else:
            return Expression(f"{self.name}[{', '.join(f'{o} + 1' for o in other)}]")

    def __str__(self):
        return self.name
    
    def _to_mz(self):
        if (self.shape is None):
            return f"{self.vtype}: {self.name} = {self.value};\n"
        
        else:
            mz_array = array_py2mz(self.value, self.shape)
            return f"array[{','.join(f'1..{d}' for d in self.shape)}] of {self.vtype}: {self.name} = {mz_array};\n"
        
        


import unittest

from test_expression import TestExpression
from test_examples import TestExamples
from test_misc import TestMisc
    

if __name__ == "__main__":
    suite = unittest.TestLoader().discover('tests')
    unittest.TextTestRunner(verbosity=1).run(suite)



from typing import List

from .exceptions import *
from .expression import *

class AnnotationConstraint:
    ANNOTATIONS = [
        ANNOTATION_BOUNDS,
        ANNOTATION_BOUNDS_Z,
        ANNOTATION_BOUNDS_R,
        ANNOTATION_BOUNDS_D,
        ANNOTATION_DOMAIN,
        ANNOTATION_VALUE_PROPAGATION,
        # Priority(k)
    ] = [
        "bounds",
        "boundsZ",
        "boundsR",
        "boundsD",
        "domain",
        "value_propagation",
    ]

class Constraint:
    CTYPES = [
        CTYPE_NORMAL,
        CTYPE_ALLDIFFERENT,
        CTYPE_AMONG,
        CTYPE_ALL_EQUAL,
        CTYPE_COUNT,
        CTYPE_INCREASING,
        CTYPE_DECREASING,
        CTYPE_DISJUNCTIVE,
        CTYPE_ARG_SORT,
    ] = [
        "normal",
        "alldifferent",
        "among",
        "all_equal",
        "count",
        "increasing",
        "decreasing",
        "disjunctive",
        "arg_sort",
    ]
    def __init__(self, cstr: ExpressionBool, ctype: str=CTYPE_NORMAL, annotation: str=None, is_redundant=False):
        self.cstr = cstr
        if (not isinstance(self.cstr, (ExpressionBool, bool, str))):
            raise PymzmValueIsNotCondition("cstr", self.cstr)

        self.ctype = ctype # This variable shouldn't be changed by the user
        if (self.ctype not in Constraint.CTYPES):
            raise PymzmInvalidConstraintType("ctype")

        self.annotation = annotation
        if (self.annotation is not None):
            if (self.annotation not in AnnotationConstraint.ANNOTATIONS):
                raise PymzmInvalidConstraintAnnotation("annotation")
            
        self.is_redundant = is_redundant

    def __str__(self):
        return self.cstr
    
    def _to_mz(self):
        if (self.is_redundant):
            return f"constraint redundant_constraint({self.cstr});\n"
        else:
            return f"constraint {self.cstr};\n"

    @staticmethod
    def _from_global_constraint(func: str, ctype: str, *args):
        return Constraint(f"{func}({', '.join(str(a) for a in args)})", ctype)

    @staticmethod
    def alldifferent(exprs: List[Expression]) -> "Constraint":
        """Constrain the elements in the passed List to be pairwise different.

        Args:
            variables (List[Expression]): Passed List of expressions

        Returns:
            Constraint: Alldifferent constraint
        """
        return Constraint._from_global_constraint("alldifferent", Constraint.CTYPE_ALLDIFFERENT, exprs)
    
    @staticmethod
    def among(n: int, exprs: List[Expression], values: List[int]):
        return Constraint._from_global_constraint("among", Constraint.CTYPE_AMONG, n, exprs, values)
    
    @staticmethod
    def all_equal(exprs: List[Expression]):
        return Constraint._from_global_constraint("all_equal", Constraint.CTYPE_ALL_EQUAL, exprs)
    
    @staticmethod
    def count(exprs: List[Expression], val: int, count: int):
        return Constraint._from_global_constraint("count", Constraint.CTYPE_COUNT, exprs, val, count)
    
    @staticmethod
    def increasing(exprs: List[Expression]):
        # Requires that the array x is in (non-strictly) increasing order (duplicates are allowed).
        return Constraint._from_global_constraint("increasing", Constraint.CTYPE_INCREASING, exprs)

    @staticmethod
    def decreasing(exprs: List[Expression]):
        # Requires that the array x is in (non-strictly) decreasing order (duplicates are allowed).
        return Constraint._from_global_constraint("decreasing", Constraint.CTYPE_DECREASING, exprs)
    
    @staticmethod
    def disjunctive(s: List[Expression], d: List[Expression]):
        # Requires that a set of tasks given by start times s and durations d do not overlap in time. 
        # Tasks with duration 0 can be scheduled at any time, even in the middle of other tasks.
        return Constraint._from_global_constraint("disjunctive", Constraint.CTYPE_DISJUNCTIVE, s, d)
    
    @staticmethod
    def disjunctive_strict(s: List[Expression], d: List[Expression]):
        # Requires that a set of tasks given by start times s and durations d do not overlap in time. 
        # Tasks with duration 0 CANNOT be scheduled at any time, but only when no other task is running.
        return Constraint._from_global_constraint("disjunctive_strict", Constraint.CTYPE_DISJUNCTIVE, s, d)
    
    @staticmethod
    def arg_sort(x: List[Expression], p: List[Expression]):
        # Constrains p to be the permutation which causes x to be in sorted order hence x[p[i]] <= x[p[i+1]].
        return Constraint._from_global_constraint("arg_sort", Constraint.CTYPE_ARG_SORT, x, p)



from .expression import *
    
def variableIterable2Str(variables) -> str:
    return str([v.name if isinstance(v, Expression) else v for v in variables]).replace("'", "")

def array_py2mz(arr, shape):
    assert len(shape) <= 2

    return str(arr).replace("], [", "|").replace("[[", "[|").replace("]]", "|]")


import unittest

import pymzm
import minizinc

class TestMisc(unittest.TestCase):
    def setUp(self):
        self.model = pymzm.Model()
        self.gecode = minizinc.Solver.lookup("gecode")

    def test_misc1(self):
        model = self.model
        # Negative summation
        xs = model.add_variables("x", range(10), pymzm.Variable.VTYPE_BOOL, 0, 1)
        ys = model.add_variables("y", range(10), pymzm.Variable.VTYPE_INTEGER, 9, 10)
        model.set_solve_criteria(pymzm.SOLVE_MAXIMIZE, pymzm.Expression.sum(xs) - pymzm.Expression.sum(ys))
        model.generate()
        
        result = minizinc.Instance(self.gecode, model).solve(all_solutions=False)

        self.assertTrue(result.solution is not None)
        self.assertTrue(result.objective < 0)





class PymzmException(Exception):
    pass

class PymzmInvalidConstraintType(PymzmException):
    def __init__(self, argname):
        self.argname = argname

    def __str__(self):
        return f"Argument \"{self.argname}\" is not a valid constraint type."

class PymzmValueIsNotCondition(PymzmException):
    def __init__(self, argname, expr):
        self.argname = argname
        self.expr = expr

    def __str__(self):
        return f"Argument \"{self.argname}\": {repr(self.expr)} (type={type(self.expr)}) is not valid as a condition."

class PymzmValueIsNotExpression(PymzmException):
    def __init__(self, argname, expr):
        self.argname = argname
        self.expr = expr

    def __str__(self):
        return f"Argument \"{self.argname}\": {repr(self.expr)} (type={type(self.expr)}) is not valid as a expression."

class PymzmNoValues(PymzmException):
    def __init__(self, argname):
        self.argname = argname

    def __str__(self):
        return f"Argument \"{self.argname}\", expected iterable but has no values."
    
class PymzmInvalidVarchoiceAnnotation(PymzmException):
    def __init__(self, argname):
        self.argname = argname

    def __str__(self):
        return f"Argument \"{self.argname}\" is not valid as a varchoice annotation."
    
class PymzmInvalidConstraintAnnotation(PymzmException):
    def __init__(self, argname):
        self.argname = argname

    def __str__(self):
        return f"Argument \"{self.argname}\" is not valid as a constraint annotation."
    
class PymzmInvalidValchoiceAnnotation(PymzmException):
    def __init__(self, argname):
        self.argname = argname

    def __str__(self):
        return f"Argument \"{self.argname}\" is not valid as a valchoice annotation."
    
class PymzmInvalidVariableError(PymzmException):
    def __init__(self, argname, msg):
        self.argname = argname
        self.msg = msg

    def __str__(self):
        return f"Argument \"{self.argname}\" error. {self.msg}."

class PymzmInvalidSearchAnnotation(PymzmException):
    def __init__(self):
        pass
    def __str__(self):
        return f"PymzmInvalidSearchAnnotation."


import unittest
import pymzm
import minizinc
import math
import sys
import os

class TestExamples(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        sys.path.append(os.path.dirname(os.path.abspath(__file__)) + "\\..")

    @classmethod
    def tearDownClass(cls):
        sys.path.pop()

    def setUp(self):
        self.model = pymzm.Model()
        self.gecode = minizinc.Solver.lookup("gecode")

    def test_australia(self):
        # https://www.minizinc.org/doc-2.5.5/en/modelling.html
        model = self.model
        nc = 3
        states = ["wa", "nsw", "nt", "v", "sa", "t", "q"]
        s = model.add_variables("state", states, val_min=1, val_max=nc)
        model.add_constraint(s["wa"] != s["nt"])
        model.add_constraint(s["wa"] != s["sa"])
        model.add_constraint(s["nt"] != s["sa"])
        model.add_constraint(s["nt"] != s["q"])
        model.add_constraint(s["sa"] != s["q"])
        model.add_constraint(s["sa"] != s["nsw"])
        model.add_constraint(s["sa"] != s["v"])
        model.add_constraint(s["q"] != s["nsw"])
        model.add_constraint(s["nsw"] != s["v"])
        model.set_solve_criteria(pymzm.SOLVE_SATISFY)
        model.generate()

        result = minizinc.Instance(self.gecode, model).solve(all_solutions=True)
        
        # Assert that solution is correct
        self.assertTrue(result.solution is not None)
        self.assertEqual(18, len(result))

    def test_integer_factorization(self):
        from examples.intfact import intfact
        n1 = 7829
        n2 = 6907
        result = intfact(self.model, self.gecode, n1, n2)

        # Assert that solution is correct
        self.assertTrue(result.solution is not None)
        self.assertEqual(result[0].x, n1)
        self.assertEqual(result[0].y, n2)

    def test_nqueens(self):
        from examples.nqueens import nqueens
        n = 8
        result = nqueens(self.model, self.gecode, n)

        # Assert that solution is correct
        self.assertTrue(result.solution is not None)
        self.assertEqual(92, len(result))

    def test_711(self):
        from examples.pr711 import pr711
        n = 4
        result = pr711(self.model, self.gecode, n)

        # Assert that solution is correct
        self.assertTrue(result.solution is not None)
        rs = [result[f"item_{i}"] / 100 for i in range(n)]
        self.assertTrue(sum(rs) == 7.11)
        self.assertTrue(math.prod(rs) == 7.11)

    def test_bibd(self):
        from examples.bibd import bibd
        # BIBD generation is described in most standard textbooks on combinatorics. 
        # A BIBD is defined as an arrangement of v  distinct objects into b blocks 
        # such that each block contains exactly k distinct objects, each object 
        # occurs in exactly r different blocks, and every two distinct objects 
        # occur together in exactly λ blocks.
        v = 7
        b = 7
        r = 3
        k = 3
        l = 1

        result = bibd(self.model, self.gecode, v, b, r, k, l)

        # Assert that solution is correct
        self.assertTrue(result.solution is not None)
        for j in range(b):
            self.assertTrue(sum(result[f"x_{i}_{j}"] for i in range(v)) == r)
        for i in range(v):
            self.assertTrue(sum(result[f"x_{i}_{j}"] for j in range(b)) == r)
        for i in range(b):
            for j in range(i):
                self.assertTrue(sum(result[f"x_{i}_{k}"] * result[f"x_{j}_{k}"] for k in range(v)) == l)

    def test_magicsquare(self):
        from examples.magicsquare import magicsquare
        n = 3
        result = magicsquare(self.model, self.gecode, n)

        # Assert that solution is correct
        y_sol = int(n * (n * n + 1) / 2)
        self.assertEqual(result["y"], y_sol)
        self.assertEqual(sum(result[f"x_{i}_{i}"] for i in range(n)), y_sol)
        self.assertEqual(sum(result[f"x_{i}_{n - 1 - i}"] for i in range(n)), y_sol)
        for i in range(n):
            self.assertEqual(sum(result[f"x_{i}_{j}"] for j in range(n)), y_sol)
        for j in range(n):
            self.assertEqual(sum(result[f"x_{i}_{j}"] for i in range(n)), y_sol)

    def test_sudoku(self):
        pass

    def test_sat(self):
        pass

if __name__ == "__main__":
    unittest.main()



from .expression import *
from .exceptions import *
from .misc import *

class ValueDict(dict):
    def __iter__(self):
        for v in self.values():
            yield v

    def __str__(self):
        return variableIterable2Str(self)
    
    
    def __invert__(self):
        raise NotImplementedError()
    def __add__(self, other):
        raise NotImplementedError()
    def __sub__(self, other):
        raise NotImplementedError()

    def __eq__(self, other: "Expression") -> "ExpressionBool":
        if (not isinstance(other, (int, float, Expression))):
            raise PymzmValueIsNotExpression("other", other)

        return [s == other for s in self]
    
    def __ne__(self, other: "Expression") -> "ExpressionBool":
        if (not isinstance(other, (int, float, Expression))):
            raise PymzmValueIsNotExpression("other", other)

        return [s != other for s in self]
    
    def __lt__(self, other: "Expression") -> "ExpressionBool":
        if (not isinstance(other, (int, float, Expression))):
            raise PymzmValueIsNotExpression("other", other)

        return [s < other for s in self]
    
    def __le__(self, other: "Expression") -> "ExpressionBool":
        if (not isinstance(other, (int, float, Expression))):
            raise PymzmValueIsNotExpression("other", other)

        return [s <= other for s in self]
    
    def __gt__(self, other: "Expression") -> "ExpressionBool":
        if (not isinstance(other, (int, float, Expression))):
            raise PymzmValueIsNotExpression("other", other)

        return [s > other for s in self]
    
    def __ge__(self, other: "Expression") -> "ExpressionBool":
        if (not isinstance(other, (int, float, Expression))):
            raise PymzmValueIsNotExpression("other", other)

        return [s >= other for s in self]
    
class Variable(Expression):
    VTYPES = [
        VTYPE_INTEGER,
        VTYPE_FLOAT,
        VTYPE_BOOL,
        VTYPE_STRING,
        VTYPE_SET,
    ] = [
        "int",
        "float",
        "bool",
        "string",
        "set"
    ]
    def __init__(self, name: str, vtype: int=VTYPE_INTEGER, val_min: int=None, val_max: int=None, domain=None):
        self.name = name
        self.vtype = vtype
        self.val_min = val_min
        self.val_max = val_max
        self.domain = domain

        if (vtype == Variable.VTYPE_INTEGER or vtype == Variable.VTYPE_SET):
            if (domain is None):
                assert self.val_min is not None
                assert self.val_max is not None

            else:
                assert self.val_min is None
                assert self.val_max is None
                #assert isinstance(domain, set)
                self.domain = set(domain)
                assert len(self.domain) > 0

        elif (vtype == Variable.VTYPE_FLOAT):
            assert self.domain is None
            assert self.val_min is not None
            assert self.val_max is not None

        elif (vtype == Variable.VTYPE_BOOL):
            assert self.val_min is None or self.val_min == 0
            assert self.val_max is None or self.val_max == 1
            self.val_min = 0
            self.val_max = 1
            self.__class__ = VariableBool
        elif (vtype == Variable.VTYPE_STRING):
            raise NotImplementedError()
        else:
            raise Exception(f"Invalid variable type vtype={vtype}")
    
    def __str__(self):
        return self.name

    def _to_mz(self):
        if (self.vtype == Variable.VTYPE_BOOL):
            return f"var bool: {self.name};\n"
        
        elif (self.vtype == Variable.VTYPE_INTEGER):
            if (self.domain is None):
                return f"var {self.val_min}..{self.val_max}: {self.name};\n"
            else:
                return f"var {self.domain}: {self.name};\n"
            
        elif (self.vtype == Variable.VTYPE_FLOAT):
            return f"var {self.val_min}..{self.val_max}: {self.name};\n"
            
        elif (self.vtype == Variable.VTYPE_SET):
            if (self.domain is None):
                return f"var set of {self.val_min}..{self.val_max}: {self.name};\n"
            else:
                return f"var set of {self.domain}: {self.name};\n"

    def __len__(self):
        assert self.vtype == Variable.VTYPE_SET
        return Expression._func("card", [self])
    
    @staticmethod
    def min(var):
        assert var.vtype == Variable.VTYPE_SET
        return Expression._func("min", [var])
    
    @staticmethod
    def max(var):
        assert var.vtype == Variable.VTYPE_SET
        return Expression._func("max", [var])
    
    def contains(self, content):
        assert self.vtype == Variable.VTYPE_SET
        return ExpressionBool(f"({content} in {self})")
    
    @staticmethod
    def intersection_length(v1, v2):
        assert v1.vtype == Variable.VTYPE_SET
        assert v2.vtype == Variable.VTYPE_SET
        return Expression(f"(card({v1} intersect {v2}))")
    
class VariableBool(ExpressionBool, Variable):
    pass